import re
from abc import ABC, abstractmethod
from collections import OrderedDict
from functools import reduce
from operator import __or__
from pprint import pformat
from types import NotImplementedType
from typing import (
    Any,
    Sequence,
    TypeVar,
    Mapping,
    Callable,
    cast,
    overload,
)
from typing import _GenericAlias as GenericAlias  # type: ignore[attr-defined]

from testfixtures import not_there, safe_pformat
from testfixtures.comparers import (
    _extract_attrs, AlreadySeen, _compare_mapping, safe_repr, compare_simple
)
from testfixtures.comparing import CompareContext, compare, register
from testfixtures.resolve import resolve, type_name
from testfixtures.utils import indent


class StatefulComparison:
    """
    A base class for stateful comparison objects.
    """

    failed: str | None = ''
    expected: Any = None
    name_attrs: Sequence[str] = ()

    def __eq__(self, other: Any) -> bool:
        return not(self != other)

    def name(self) -> str:
        name = type(self).__name__
        if self.name_attrs:
            name += '(%s)' % ', '.join('%s=%r' % (n, getattr(self, n)) for n in self.name_attrs)
        return name

    def body(self) -> str:
        return safe_pformat(self.expected)[1:-1]

    def __repr__(self) -> str:
        name = self.name()
        body = self.failed or self.body()
        prefix = '<%s%s>' % (name, self.failed and '(failed)' or '')
        if '\n' in body:
            return '\n'+prefix+'\n'+body.strip('\n')+'\n'+'</%s>' % name
        elif body:
            return prefix + body + '</>'
        return prefix


class Comparison(StatefulComparison):
    """
    These are used when you need to compare an object's type, a subset of its attributes
    or make equality checks with objects that do not natively support comparison.

    :param object_or_type: The object or class from which to create the
                           :class:`Comparison`.

    :param attribute_dict: An optional dictionary containing attributes
                           to place on the :class:`Comparison`.

    :param partial:
      If true, only the specified attributes will be checked and any extra attributes
      of the object being compared with will be ignored.

    :param attributes: Any other keyword parameters passed will placed
                       as attributes on the :class:`Comparison`.

    """

    def __init__(self,
                 object_or_type: Any,
                 attribute_dict: dict[str, Any] | None = None,
                 partial: bool = False,
                 **attributes: Any):
        self.partial = partial
        if attributes:
            if attribute_dict is None:
                attribute_dict = attributes
            else:
                attribute_dict.update(attributes)
        if isinstance(object_or_type, str):
            c = resolve(object_or_type).found
            if c is not_there:
                raise AttributeError(
                    '%r could not be resolved' % object_or_type
                )
        elif isinstance(object_or_type, type):
            c = object_or_type
        else:
            c = object_or_type.__class__
            if attribute_dict is None:
                attribute_dict = _extract_attrs(object_or_type)
        self.expected_type = c
        self.expected_attributes = attribute_dict

    def __ne__(self, other: Any) -> bool:

        if isinstance(other, AlreadySeen):
            other = other.obj

        other_type = type(other)
        if self.expected_type is not other_type:
            self.failed = 'wrong type: ' + type_name(other_type)
            return True

        if self.expected_attributes is None:
            return False

        attribute_names = set(self.expected_attributes.keys())
        actual_attributes: dict[str, Any]
        if self.partial:
            actual_attributes = {}
        else:
            actual_attributes = cast(dict[str, Any], _extract_attrs(other))
            attribute_names -= set(actual_attributes)

        for name in attribute_names:
            try:
                actual_attributes[name] = getattr(other, name)
            except AttributeError:
                pass

        context = CompareContext(x_label='Comparison', y_label='actual')
        self.failed = _compare_mapping(self.expected_attributes,
                                       actual_attributes,
                                       context,
                                       obj_for_class=not_there,
                                       prefix='attributes ',
                                       breadcrumb='.%s',
                                       check_y_not_x=not self.partial)
        return bool(self.failed)

    def name(self) -> str:
        return 'C:' + type_name(self.expected_type)

    def body(self) -> str:
        if self.expected_attributes:
            # if we're not failed, show what we will expect:
            lines = []
            for k, v in sorted(self.expected_attributes.items()):
                rv = safe_repr(v)
                if '\n' in rv:
                    rv = indent(rv)
                lines.append(f'{k}: {rv}')
            return '\n'.join(lines)
        return ''


class SequenceComparison(StatefulComparison):
    """
    An object that can be used in comparisons of expected and actual
    sequences.

    :param expected: The items expected to be in the sequence.
    :param ordered:
      If ``True``, then the items are expected to be in the order specified.
      If ``False``, they may be in any order.
      Defaults to ``True``.
    :param partial:
      If ``True``, then any keys not expected will be ignored.
      Defaults to ``False``.
    :param recursive:
      If a difference is found, recursively compare the item where
      the difference was found to highlight exactly what was different.
      Defaults to ``False``.
    """

    name_attrs: Sequence[str] = ('ordered', 'partial')

    def __init__(
            self,
            *expected: Any,
            ordered: bool = True,
            partial: bool = False,
            recursive: bool = False,
    ):
        self.expected = expected
        self.ordered = ordered
        self.partial = partial
        self.recursive = recursive
        self.checked_indices = set[int]()

    def __ne__(self, other: Any) -> bool:
        actual: list[Any]
        try:
            actual = original_actual = list(other)
        except TypeError:
            self.failed = 'bad type'
            return True
        expected = list(self.expected)
        actual = list(actual)

        matched = []
        matched_expected_indices = []
        matched_actual_indices = []

        missing_from_expected = actual
        missing_from_expected_indices = actual_indices = list(range(len(actual)))

        missing_from_actual = []
        missing_from_actual_indices = []

        start = 0
        for e_i, e in enumerate(expected):
            try:
                i = actual.index(e, start)
                a_i = actual_indices.pop(i)
            except ValueError:
                missing_from_actual.append(e)
                missing_from_actual_indices.append(e_i)
            else:
                matched.append(missing_from_expected.pop(i))
                matched_expected_indices.append(e_i)
                matched_actual_indices.append(a_i)
                self.checked_indices.add(a_i)
                if self.ordered:
                    start = i

        matches_in_order = matched_actual_indices == sorted(matched_actual_indices)
        all_matched = not (missing_from_actual or missing_from_expected)
        partial_match = self.partial and not missing_from_actual

        if (matches_in_order or not self.ordered) and (all_matched or partial_match):
            return False

        expected_indices = matched_expected_indices+missing_from_actual_indices
        actual_indices = matched_actual_indices

        if self.partial:
            # try to give a clue as to what didn't match:
            if self.recursive and self.ordered and missing_from_expected:
                actual_indices.append(missing_from_expected_indices.pop(0))
                missing_from_expected.pop(0)

            ignored = missing_from_expected
            missing_from_expected = []
        else:
            actual_indices += missing_from_expected_indices
            ignored = None

        message = []

        def add_section(name: str, content: Any) -> None:
            if content:
                message.append(name+':\n'+pformat(content))

        add_section('ignored', ignored)

        if self.ordered:
            message.append(cast(str, compare(
                expected=[self.expected[i] for i in sorted(expected_indices)],
                actual=[original_actual[i] for i in sorted(actual_indices)],
                recursive=self.recursive,
                raises=False
            )).split('\n\n', 1)[1])
        else:
            add_section('same', matched)
            add_section('in expected but not actual', missing_from_actual)
            add_section('in actual but not expected', missing_from_expected)

        self.failed = '\n\n'.join(message)
        return True


class Subset(SequenceComparison):
    """
    A shortcut for :class:`SequenceComparison` that checks if the
    specified items are present in the sequence.
    """

    name_attrs = ()

    def __init__(self, *expected: Any) -> None:
        super(Subset, self).__init__(*expected, ordered=False, partial=True)


class Permutation(SequenceComparison):
    """
    A shortcut for :class:`SequenceComparison` that checks if the set of items
    in the sequence is as expected, but without checking ordering.
    """

    def __init__(self, *expected: Any) -> None:
        super(Permutation, self).__init__(*expected, ordered=False, partial=False)


class MappingComparison(StatefulComparison):
    """
    An object that can be used in comparisons of expected and actual
    mappings.

    :param expected_mapping:
      The mapping that should be matched expressed as either a sequence of
      ``(key, value)`` tuples or a mapping.
    :param expected_items: The items that should be matched.
    :param ordered:
      If ``True``, then the keys in the mapping are expected to be in the order specified.
      Defaults to ``False``.
    :param partial:
      If ``True``, then any keys not expected will be ignored.
      Defaults to ``False``.
    :param recursive:
      If a difference is found, recursively compare the value where
      the difference was found to highlight exactly what was different.
      Defaults to ``False``.
    """

    name_attrs = ('ordered', 'partial')

    def __init__(
            self,
            *expected_mapping: tuple[Any, Any] | Mapping[Any, Any],
            ordered: bool = False,
            partial: bool = False,
            recursive: bool = False,
            **expected_items: Any,
    ):
        self.ordered = ordered
        self.partial = partial
        self.recursive = recursive

        if len(expected_mapping) == 1:
            expected = OrderedDict(*expected_mapping)
        else:
            expected = OrderedDict(expected_mapping)  # type: ignore[arg-type]
        expected.update(expected_items)

        self.expected = expected

    def body(self) -> str:
        parts = []
        text_length = 0
        for key, value in self.expected.items():
            part = f'{safe_repr(key)}: {safe_pformat(value)}'
            text_length += len(part)
            parts.append(part)
        if text_length > 60:
            sep = ',\n'
        else:
            sep = ', '
        return sep.join(parts)

    def __ne__(self, other: Any) -> bool:
        try:
            actual_keys = other.keys()
            actual_mapping = dict(other.items())
        except AttributeError:
            self.failed = 'bad type'
            return True

        expected_keys = self.expected.keys()
        expected_mapping = self.expected

        if self.partial:
            ignored_keys = set(actual_keys) - set(expected_keys)
            for key in ignored_keys:
                del actual_mapping[key]
            # preserve the order:
            actual_keys = [k for k in actual_keys if k not in ignored_keys]
        else:
            ignored_keys = None

        mapping_differences = compare(
            expected=expected_mapping,
            actual=actual_mapping,
            recursive=self.recursive,
            raises=False
        )

        if self.ordered:
            key_differences = compare(
                expected=list(expected_keys),
                actual=list(actual_keys),
                recursive=self.recursive,
                raises=False
            )
        else:
            key_differences = None

        if key_differences or mapping_differences:

            message = []

            if ignored_keys:
                message.append('ignored:\n'+pformat(sorted(ignored_keys)))

            if mapping_differences:
                message.append(mapping_differences.split('\n\n', 1)[1])

            if key_differences:
                message.append('wrong key order:\n\n'+key_differences.split('\n\n', 1)[1])

            self.failed = '\n\n'.join(message)

            return True
        return False


class TextComparison:
    """
    An object that can be used in comparisons of expected and actual
    strings where the string expected matches a pattern rather than a
    specific concrete string.

    :param regex_source: A string containing the source for a regular
                         expression that will be used whenever this
                         :class:`TextComparison` is compared with
                         any :class:`str` instance, or an already compiled
                         :class:`re.Pattern`.

    :param flags: Flags passed to :func:`re.compile`.

    :param flag_names: See the :ref:`examples <textcomparison>`.
    """

    @overload
    def __init__(self, regex_source: re.Pattern[str]): ...

    @overload
    def __init__(self, regex_source: str, flags: int | None = None, **flag_names: bool): ...

    def __init__(
        self,
        regex_source: str | re.Pattern[str],
        flags: int | None = None,
        **flag_names: bool,
    ):
        if isinstance(regex_source, re.Pattern):
            self.re = regex_source
            return

        args: list[Any] = [regex_source]

        flags_ = []
        if flags:
            flags_.append(flags)
        flags_.extend(getattr(re, f.upper()) for f in flag_names)
        if flags_:
            args.append(reduce(__or__, flags_))

        self.re = re.compile(*args)

    def __eq__(self, other: Any) -> bool | NotImplementedType:
        if not isinstance(other, str):
            return NotImplemented
        return self.re.match(other) is not None

    def __repr__(self) -> str:
        return '<S:%s>' % self.re.pattern

    def __lt__(self, other: Any) -> bool:
        return self.re.pattern < other

    def __gt__(self, other: Any) -> bool:
        return self.re.pattern > other


# Deprecated alias for TextComparison. Note this is NOT the same as
# StrComparison, which checks an object's type and its str().
StringComparison = TextComparison


class RoundComparison:
    """
    An object that can be used in comparisons of expected and actual
    numerics to a specified precision.

    :param value: numeric to be compared.

    :param precision: Number of decimal places to round to in order
                      to perform the comparison.
    """
    def __init__(self, value: float, precision: int):
        self.rounded = round(value, precision)
        self.precision = precision

    def __eq__(self, other: Any) -> bool:
        other_rounded = round(other, self.precision)
        if type(self.rounded) is not type(other_rounded):
            raise TypeError('Cannot compare %r with %r' % (self, type(other)))
        return self.rounded == other_rounded

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return '<R:%s to %i digits>' % (self.rounded, self.precision)


class RangeComparison:
    """
    An object that can be used in comparisons of orderable types to
    check that a value specified within the given range.

    :param lower_bound: the inclusive lower bound for the acceptable range.

    :param upper_bound: the inclusive upper bound for the acceptable range.
    """
    def __init__(self, lower_bound: Any, upper_bound: Any) -> None:
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __eq__(self, other: Any) -> bool:
        return self.lower_bound <= other <= self.upper_bound

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return '<Range: [%s, %s]>' % (self.lower_bound, self.upper_bound)


class RenderingComparison(ABC):
    """
    A base for comparisons that check an object's type and a string rendering
    of it produced by :attr:`render`.
    """

    @staticmethod
    @abstractmethod
    def render(other: object, /) -> str:
        """The callable used to render the compared object as a string."""

    @overload
    def __init__(self, type_: type, expected: str): ...
    @overload
    def __init__(self, type_: type, *, match: str | re.Pattern[str]): ...

    def __init__(
        self,
        type_: type,
        expected: str | None = None,
        *,
        match: str | re.Pattern[str] | None = None,
    ):
        if (expected is None) == (match is None):
            raise TypeError('provide either expected or match')
        self.type = type_
        self.expected = expected
        self.match = match

    def __eq__(self, other: Any) -> bool:
        if type(other) is not self.type:
            return False
        rendered = self.render(other)
        if self.match is not None:
            return re.search(self.match, rendered) is not None
        return rendered == self.expected

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        detail = self.expected if self.expected is not None else f'match={self.match!r}'
        return f'<{type(self).__name__}: {type_name(self.type)}: {detail}>'


class ReprComparison(RenderingComparison):
    """
    An object that can be used in comparisons to check that an object is both
    of an expected type and has an expected :func:`repr`.

    :param type_: the type the compared object must be exactly; subclasses do
                  not match.

    :param expected: the :func:`repr` the compared object must have exactly.

    :param match: a regular expression, as either a :class:`str` or a compiled
                  :class:`re.Pattern`, that the compared object's :func:`repr`
                  must match. Mutually exclusive with ``expected``.
    """
    render = staticmethod(repr)


class StrComparison(RenderingComparison):
    """
    An object that can be used in comparisons to check that an object is both
    of an expected type and has an expected :class:`str`.

    :param type_: the type the compared object must be exactly; subclasses do
                  not match.

    :param expected: the :class:`str` the compared object must have exactly.

    :param match: a regular expression, as either a :class:`str` or a compiled
                  :class:`re.Pattern`, that the compared object's :class:`str`
                  must match. Mutually exclusive with ``expected``.
    """
    render = staticmethod(str)


T = TypeVar('T')

@overload
def like(t: str | re.Pattern[str]) -> str: ...


@overload
def like(t: type[T], **attributes: Any) -> T: ...


def like(t: type[T] | str | re.Pattern[str], **attributes: Any) -> T | str:
    """
    Create a type-safe partial comparison for use in strictly typed code.

    This is a convenience function that creates a :class:`Comparison` with
    ``partial=True`` but is typed to return the type being compared, making it
    compatible with strict type checkers like mypy.

    If passed a :class:`str` pattern or a compiled :class:`re.Pattern`, a
    :class:`TextComparison` typed as a :class:`str` is returned instead.

    :param t: The type to compare against, or a regular expression pattern.
    :param attributes: Keyword arguments specifying the attributes to check.
    :return: A :class:`Comparison` typed as the input type, or a
             :class:`TextComparison` typed as a :class:`str`.
    """
    if isinstance(t, (str, re.Pattern)):
        return cast(str, TextComparison(t))
    return Comparison(t, attribute_dict=attributes, partial=True)  # type: ignore[return-value]


@overload
def repr_like(type_: type[T], expected: str) -> T: ...


@overload
def repr_like(type_: type[T], *, match: str | re.Pattern[str]) -> T: ...


def repr_like(
    type_: type[T],
    expected: str | None = None,
    *,
    match: str | re.Pattern[str] | None = None,
) -> T:
    """
    Create a type-safe :class:`ReprComparison` for use in strictly typed code.

    This is a convenience function that creates a :class:`ReprComparison` but is
    typed to return the type being compared, making it compatible with strict
    type checkers like mypy.

    :param type_: the type the compared object must be exactly; subclasses do
                  not match.
    :param expected: the :func:`repr` the compared object must have exactly.
    :param match: a regular expression, as either a :class:`str` or a compiled
                  :class:`re.Pattern`, that the compared object's :func:`repr`
                  must match. Mutually exclusive with ``expected``.
    :return: A :class:`ReprComparison` typed as the input type.
    """
    return cast(T, ReprComparison(type_, expected, match=match))  # type: ignore[call-overload]


@overload
def str_like(type_: type[T], expected: str) -> T: ...


@overload
def str_like(type_: type[T], *, match: str | re.Pattern[str]) -> T: ...


def str_like(
    type_: type[T],
    expected: str | None = None,
    *,
    match: str | re.Pattern[str] | None = None,
) -> T:
    """
    Create a type-safe :class:`StrComparison` for use in strictly typed code.

    This is a convenience function that creates a :class:`StrComparison` but is
    typed to return the type being compared, making it compatible with strict
    type checkers like mypy.

    :param type_: the type the compared object must be exactly; subclasses do
                  not match.
    :param expected: the :class:`str` the compared object must have exactly.
    :param match: a regular expression, as either a :class:`str` or a compiled
                  :class:`re.Pattern`, that the compared object's :class:`str`
                  must match. Mutually exclusive with ``expected``.
    :return: A :class:`StrComparison` typed as the input type.
    """
    return cast(T, StrComparison(type_, expected, match=match))  # type: ignore[call-overload]


S = TypeVar("S", bound=Sequence[Any])
S_ = TypeVar("S_", bound=Sequence[Any])


@overload
def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
) -> Callable[[S], S]: ...


@overload
def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
    *,
    returns: type[S_],
) -> Callable[[S], S_]: ...


def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
    *,
    returns: type[S_] | None = None,
) -> Callable[[S], S | S_]:
    """
    Create a type-safe sequence comparison with configurable partial matching
    and ordering requirements.

    This function returns a callable that wraps a sequence in a comparison object,
    making it compatible with strict type checkers.

    :param partial: If ``True``, only items in the expected sequence need to be present
                    in the actual sequence. Defaults to ``False``.
    :param ordered: If ``True``, items must appear in the same order. Defaults to ``True``.
    :param recursive: If ``True``, provide detailed recursive comparison when differences
                      are found. Defaults to ``True``.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A callable that takes a sequence and returns a comparison object typed
             as a sequence.
    """
    def maker(items: S) -> S | S_:
        return SequenceComparison(  # type: ignore[return-value]
            *items, partial=partial, ordered=ordered, recursive=recursive
        )

    return maker


@overload
def contains(
    items: S,
) -> S: ...


@overload
def contains(
    items: S,
    *,
    returns: type[S_],
) -> S_: ...


def contains(
    items: S,
    *,
    returns: type[S_] | None = None,
) -> S | S_:
    """
    Create a type-safe partial sequence comparison that ignores order.

    Checks that the specified items are present in the actual sequence, regardless
    of their order or what other items are present. This is useful when you only
    care that certain elements exist in a collection.

    :param items: The sequence of items that must be present.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A comparison object typed as a sequence.
    """
    return SequenceComparison(  # type: ignore[return-value]
        *items, ordered=False, partial=True, recursive=True
    )


@overload
def unordered(
    items: S,
) -> S: ...


@overload
def unordered(
    items: S,
    *,
    returns: type[S_],
) -> S_: ...


def unordered(
    items: S,
    *,
    returns: type[S_] | None = None,
) -> S | S_:
    """
    Create a type-safe sequence comparison that ignores order but requires all
    items to match.

    Checks that the actual sequence contains exactly the same items as specified,
    but in any order. This is useful when order doesn't matter but you want to
    ensure no extra or missing items.

    :param items: The sequence of items that must match exactly.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A comparison object typed as a sequence.
    """
    return SequenceComparison(  # type: ignore[return-value]
        *items, ordered=False, partial=False, recursive=True
    )


M = TypeVar("M", bound=Mapping[Any, Any])
M_ = TypeVar("M_", bound=Mapping[Any, Any])


@overload
def mapping(
    partial: bool = False,
    ordered: bool = False,
    recursive: bool = True,
) -> Callable[[M], M]: ...


@overload
def mapping(
    partial: bool = False,
    ordered: bool = False,
    recursive: bool = True,
    *,
    returns: type[M_],
) -> Callable[[M], M_]: ...


def mapping(
    partial: bool = False,
    ordered: bool = False,
    recursive: bool = True,
    *,
    returns: type[M_] | None = None,
) -> Callable[[M], M | M_]:
    """
    Create a type-safe mapping comparison with configurable key ordering and
    partial matching.

    This is the mapping equivalent of :func:`sequence`. It returns a callable
    that wraps a mapping in a :class:`MappingComparison`, typed to match the
    mapping being compared so that strict type checkers stay happy.

    :param partial: If ``True``, any keys not expected are ignored.
                    Defaults to ``False``.
    :param ordered: If ``True``, the keys must appear in the order given.
                    Defaults to ``False``.
    :param recursive: If ``True``, provide detailed recursive comparison when
                      differences are found. Defaults to ``True``.
    :param returns: Optional type hint for the return type, used to satisfy type
                    checkers when the comparison needs to appear as a different
                    mapping type.
    :return: A callable that takes a mapping and returns a comparison object
             typed as a mapping.
    """
    def maker(items: M) -> M | M_:
        return MappingComparison(  # type: ignore[return-value]
            items, ordered=ordered, partial=partial, recursive=recursive
        )

    return maker


register(StatefulComparison, compare_simple)



"""
Tools for helping to test applications that use Pydantic.
"""
from typing import TYPE_CHECKING

import pydantic as pydantic
from pydantic import BaseModel

from .comparers import _compare_mapping, compare_simple

if TYPE_CHECKING:
    from .comparing import CompareContext


def compare_basemodel(
        x: BaseModel,
        y: BaseModel,
        context: 'CompareContext',
) -> str | None:
    """
    Returns an informative string describing the differences between the two
    supplied Pydantic :class:`~pydantic.BaseModel` instances, based on their
    declared fields.
    """
    if type(x) is not type(y):
        return compare_simple(x, y, context)
    x_attrs = x.__dict__.copy()
    y_attrs = y.__dict__.copy()
    if not context.qualified_equals(x_attrs, y_attrs):
        return _compare_mapping(x_attrs, y_attrs, context, x, 'attributes ', '.%s')
    return None



"""
Tools for helping to test applications that use NumPy.
"""
from typing import TYPE_CHECKING

import numpy as numpy
from numpy import ndarray
from numpy.ma import MaskedArray, getmaskarray
from numpy.testing import assert_allclose, assert_array_equal

if TYPE_CHECKING:
    from .comparing import CompareContext

INEXACT_KINDS = 'fc'
# The tolerances used by the pandas and polars frame comparers, rather than
# assert_allclose's tighter defaults, so float semantics agree across all three:
RTOL = 1e-5
ATOL = 1e-8


def compare_ndarray(x: ndarray, y: ndarray, context: 'CompareContext') -> str | None:
    """
    Returns a textual description of the differences between two
    :class:`numpy.ndarray` instances, as reported by
    :func:`numpy.testing.assert_allclose` for float and complex arrays
    or :func:`numpy.testing.assert_array_equal` for arrays of any other
    :class:`dtype <numpy.dtype>`.

    Shape and dtype must always match: there is no broadcasting and an
    ``int32`` array is never equal to an ``int64`` one. NaNs in matching
    positions compare equal.

    When ``strict=True`` is passed to :func:`~testfixtures.compare`,
    float and complex values must be exactly equal; otherwise ``rtol=1e-5``
    and ``atol=1e-8`` apply, matching the pandas and polars comparers.
    """
    try:
        if x.dtype.kind in INEXACT_KINDS and not context.strict:
            assert_allclose(y, x, rtol=RTOL, atol=ATOL, strict=True)
        else:
            assert_array_equal(y, x, strict=True)
    except AssertionError as e:
        return str(e)
    return None


def compare_masked_array(x: MaskedArray, y: MaskedArray, context: 'CompareContext') -> str | None:
    """
    Returns a textual description of the differences between two
    :class:`numpy.ma.MaskedArray` instances.

    Masks are part of the data, so they must match exactly regardless of
    other options; data under masked positions is ignored. The unmasked
    data is then compared as described for
    :func:`~testfixtures.numpy.compare_ndarray`.
    """
    masks = compare_ndarray(getmaskarray(x), getmaskarray(y), context)
    if masks is not None:
        return 'masks differ: ' + masks
    return compare_ndarray(x.compressed(), y.compressed(), context)



import atexit
import logging
import warnings
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from logging import LogRecord
from pprint import pformat
from types import TracebackType
from typing import (
    Any, Callable, Iterator, Sequence, TypeAlias, TypeVar, List, Tuple, Self, Protocol, overload
)
from warnings import warn

from .comparing import compare
from .comparison import SequenceComparison
from .utils import wrap


class CaptureSource(Protocol):
    """
    Protocol that :class:`LogCapture` sources must implement.
    """

    def install(self, collector: Callable[['Entry'], None]) -> None:
        """
        Start capturing.

        :param collector: The destination for captured :attr:`~testfixtures.LogCapture.entries`.
        """
        ...

    def uninstall(self) -> None:
        """Stop capturing and restore any previous state."""
        ...



T = TypeVar('T')
AttributeSpec: TypeAlias = Sequence[str | Callable[[T], Any]] | Callable[[T], Any]


def build_actual_extractor(
    attributes: AttributeSpec[T],
    extract_field: Callable[[T, str], Any],
) -> Callable[[T], Any]:
    # Compile the attributes spec into a callable that extracts an entry's
    # ``actual`` value from a raw record. A bare string is treated as a single
    # attribute name (``str`` is itself a ``Sequence[str]``).
    if callable(attributes):
        return attributes
    if isinstance(attributes, str):
        attributes = (attributes,)
    attrs = tuple(attributes)
    if len(attrs) == 1:
        only = attrs[0]
        if callable(only):
            return only
        return lambda raw: extract_field(raw, only)
    return lambda raw: tuple(
        a(raw) if callable(a) else extract_field(raw, a) for a in attrs
    )


@dataclass
class Entry:
    """
    A captured log entry captured by a :class:`~testfixtures.logcapture.CaptureSource`.
    """

    #: The raw object delivered by the logging framework. This is a :class:`~logging.LogRecord`
    #: for standard library logging.
    raw: Any
    #: The extracted value used by :meth:`~testfixtures.LogCapture.check` and related methods.
    #: Its structure is determined by the ``attributes`` parameter of the source.
    actual: Any
    #: Numeric log level used by :meth:`~testfixtures.LogCapture.ensure_checked`, or ``None``
    #: if the source does not provide a comparable numeric level.
    level: int | None
    #: The exception associated with this entry if one was captured, otherwise ``None``.
    exception: BaseException | None = None
    #: Whether this entry has been marked as checked by a :meth:`~testfixtures.LogCapture.check`
    #: call.
    checked: bool = False

LogRecordAttributes: TypeAlias = AttributeSpec[LogRecord]


class LogCapture:
    """
    Captures log entries from one or more sources and provides methods to check what was logged.

    :param sources:
        One or more :class:`capture sources <testfixtures.logcapture.CaptureSource>`
    :param install:
        If ``True`` (the default), capturing starts immediately on construction.
        If ``False``, installation is deferred until :meth:`install` is called.
    :param recursive_check:
        If ``True``, entries are compared recursively by :meth:`check`.
    :param ensure_checks_above:
        The log level above which entries must have been checked.
        See :meth:`ensure_checked`.

    For compatibility with earlier versions, capturing only
    standard library :mod:`logging` is supported by instantiating using these parameters:

    :param names:
        A string (or tuple of strings) containing the dotted name(s) of loggers to capture.
        By default, the root logger is captured.
    :param level:
        The minimum log level to capture.  Defaults to ``1``, capturing everything.
    :param propagate:
        If specified, any captured loggers will have their propagate attribute set to the supplied
        value. This can be used to prevent propagation from a child logger to a parent logger that
        has configured handlers.
    :param attributes:
        The sequence of attributes to return for each record or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, an attribute of that name will
        be obtained from the :class:`logging.LogRecord`. If the attribute is callable, the result of
        calling it will be used as the attribute value. If an attribute is missing, ``None`` will be
        used in its place.

        If a callable is passed, it will be called with the :class:`~logging.LogRecord` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.
    """

    #: The list of :class:`entries <testfixtures.logcapture.Entry>` captured so far.
    entries: list[Entry]

    @property
    def records(self) -> List[LogRecord]:
        """
        The records captured by this :class:`LogCapture`.

        .. deprecated:: 12
            Use the :attr:`entries` attribute instead.
        """
        return [e.raw for e in self.entries if isinstance(e.raw, LogRecord)]


    instances: set['LogCapture'] = set()
    atexit_setup = False
    installed = False

    ensure_checks_above: int | None
    #: Class-level default for :meth:`ensure_checked`. Set this to apply a level
    #: threshold to all :class:`LogCapture` instances that do not override it explicitly.
    default_ensure_checks_above: int | None = None

    @overload
    def __init__(
        self,
        *sources: CaptureSource,
        install: bool = True,
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        names: str | Tuple[str | None, ...] | None = None,
        *,
        install: bool = True,
        level: int = 1,
        propagate: bool | None = None,
        attributes: LogRecordAttributes = ...,
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None: ...

    def __init__(  # type: ignore[misc]
        self,
        *args: Any,
        install: bool = True,
        level: int = 1,
        propagate: bool | None = None,
        attributes: LogRecordAttributes = (
            'name',
            'levelname',
            'getMessage',
        ),
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None:
        if args and hasattr(args[0], 'install'):
            self._sources = list(args)
        else:
            names: str | Tuple[str | None, ...] | None = args[0] if args else None
            if not isinstance(names, tuple):
                names = (names,)
            self._sources = [LoggingSource(attributes, level, names=names, propagate=propagate)]
        self.recursive_check = recursive_check
        self._disabled = False
        if ensure_checks_above is None:
            self.ensure_checks_above = self.default_ensure_checks_above
        else:
            self.ensure_checks_above = ensure_checks_above
        self.clear()
        if install:
            self.install()

    @classmethod
    def atexit(cls) -> None:
        if cls.instances:
            warnings.warn(
                'LogCapture instances not uninstalled by shutdown, '
                'sources captured:\n'
                '%s' % ('\n'.join((', '.join(repr(s) for s in i._sources) for i in cls.instances)))
            )

    def __len__(self) -> int:
        return len(self.entries)

    def __getitem__(self, index: int) -> Any:
        return self.entries[index].actual

    def __contains__(self, what: Any) -> bool:
        for i, entry in enumerate(self.entries):
            if what == entry.actual:
                self.entries[i].checked = True
                return True
        return False

    def clear(self) -> None:
        """Clear any entries that have been captured."""
        self.entries: list[Entry] = []

    def mark_all_checked(self) -> None:
        """
        Mark all captured events as checked.
        This should be called if you have made assertions about logging
        other than through :class:`LogCapture` methods.
        """
        for entry in self.entries:
            entry.checked = True

    def ensure_checked(self, level: int | None = None) -> None:
        """
        Ensure every entry logged above the specified `level` has been checked.
        Raises an :class:`AssertionError` if this is not the case.

        :param level:
           This defaults to the level passed during :class:`LogCapture` instantiation.
        """
        threshold: int | None = level if level is not None else self.ensure_checks_above
        if threshold is None:
            return
        un_checked = []
        for entry in self.entries:
            if entry.level is not None and entry.level >= threshold and not entry.checked:
                un_checked.append(entry.actual)
        if un_checked:
            raise AssertionError(('Not asserted ERROR log(s): %s') % (pformat(un_checked)))

    def _collect_entry(self, entry: Entry) -> None:
        self.entries.append(entry)

    @contextmanager
    def disabled(self) -> Iterator[None]:
        """
        A context manager that stops capturing for the duration of the ``with`` block by
        uninstalling the capture and reinstalling it on exit. Anything logged while it is
        active is handled as if the capture were not present, which is useful for ignoring
        noisy setup in a test.
        """
        self.uninstall()
        try:
            yield
        finally:
            self.install()

    def install(self) -> Self | None:
        """
        Install this :class:`LogCapture`, enabling all configured sources to begin capturing.
        """
        for source in self._sources:
            source.install(self._collect_entry)
        self.instances.add(self)
        if not self.__class__.atexit_setup:
            atexit.register(self.atexit)
            self.__class__.atexit_setup = True
        return None

    def uninstall(self) -> None:
        """
        Uninstall this :class:`LogCapture`, restoring all sources to their previous state.
        """
        if self in self.instances:
            for source in self._sources:
                source.uninstall()
            self.instances.remove(self)

    @classmethod
    def uninstall_all(cls) -> None:
        "This will uninstall all existing :class:`LogCapture` objects."
        for i in tuple(cls.instances):
            i.uninstall()

    def actual(self) -> list[Any]:
        """
        The sequence of :attr:`~testfixtures.logcapture.Entry.actual` items captured.

        This can be useful for making more complex assertions about captured logging.
        The full :class:`~testfixtures.logcapture.Entry` objects captured for each logging call
        can be accessed through :attr:`entries`.
        """
        return [e.actual for e in self.entries]

    def __str__(self) -> str:
        if not self.entries:
            return 'No logging captured'
        tuples = (r if isinstance(r, tuple) else (r,) for r in self.actual())
        return '\n'.join(' '.join(str(e) for e in t) for t in tuples)

    def check(
        self,
        *expected: Any,
        order_matters: bool = True,
        raises: bool = True,
        level: int | None = None,
        predicate: Callable[[Entry], bool] | None = None,
    ) -> str | None:
        """
        This will compare the captured entries with the expected
        entries provided and raise an :class:`AssertionError` if they
        do not match.

        :param expected:

          A sequence of expected :attr:`~testfixtures.logcapture.Entry.actual` items.

        :param order_matters:

          A keyword-only parameter that controls whether the order of the
          captured entries is required to match those of the expected entries.
          Defaults to ``True``.

        :param raises: If ``False``, the message that would be raised in the
                       :class:`AssertionError` will be returned instead of the
                       exception being raised.

        :param level:

          A keyword-only parameter that, if provided, excludes any captured entry
          with a numeric :attr:`~testfixtures.logcapture.Entry.level` below it from
          the comparison. Entries with a ``level`` of ``None`` are always included.

        :param predicate:

          A keyword-only parameter that, if provided, is called with each captured
          :class:`~testfixtures.logcapture.Entry`; only those for which it returns
          a true value are included in the comparison.
        """
        __tracebackhide__ = True

        actual = []
        for entry in self.entries:
            if level is not None and entry.level is not None and entry.level < level:
                continue
            if predicate is not None and not predicate(entry):
                continue
            entry.checked = True
            actual.append(entry.actual)

        result = None
        if order_matters:
            result = compare(
                expected, actual=actual, recursive=self.recursive_check, raises=False
            )
        else:
            expected_ = SequenceComparison(
                *expected, ordered=False, partial=False, recursive=self.recursive_check
            )
            if expected_ != actual:
                result = expected_.failed
        if result and raises:
            raise AssertionError(result)
        return result

    def check_empty(
        self,
        level: int | None = None,
        predicate: Callable[[Entry], bool] | None = None,
    ) -> None:
        """
        Assert that nothing was logged, raising an :class:`AssertionError` if it was.
        This is a clearer way of writing :meth:`check` with no expected entries.

        :param level:

          If provided, only entries with a numeric
          :attr:`~testfixtures.logcapture.Entry.level` at or above it are considered.
          Entries with a ``level`` of ``None`` are always considered.

        :param predicate:

          If provided, it is called with each captured
          :class:`~testfixtures.logcapture.Entry`; only those for which it returns
          a true value are considered.
        """
        __tracebackhide__ = True
        self.check(level=level, predicate=predicate)

    def raise_first_exception(self, start_index: int = 0) -> None:
        """
        Raise the first captured exception from ``start_index`` onwards.

        :param start_index: The index into :attr:`entries` from where to start looking.
        """
        for entry in self.entries[start_index:]:
            if entry.exception is not None:
                raise entry.exception

    def check_exception_str(self, expected: str, index: int = -1) -> None:
        """
        Check the string representation of a captured exception.

        :param expected: The expected string.
        :param index: The index into :attr:`entries` where the exception should have been captured.
        """
        compare(expected, actual=str(self.entries[index].exception))

    def check_present(
            self, *expected: Any, order_matters: bool = True, raises: bool = True
    ) -> str | None:
        """
        This will check if the captured attr:`entries` contain all of the expected
        entries provided and raise an :class:`AssertionError` if not.
        This will ignore entries that have been captured but that do not
        match those in ``expected``.

        :param expected:

          A sequence of expected :attr:`~testfixtures.logcapture.Entry.actual` items.

        :param order_matters:

          A keyword-only parameter that controls whether the order of the
          captured entries is required to match those of the expected entries.
          Defaults to ``True``.

        :param raises: If ``False``, the message that would be raised in the
                       :class:`AssertionError` will be returned instead of the
                       exception being raised.
        """
        __tracebackhide__ = True
        actual = self.actual()
        expected_ = SequenceComparison(
            *expected, ordered=order_matters, partial=True, recursive=self.recursive_check
        )
        if expected_ != actual:
            if raises:
                raise AssertionError(expected_.failed)
            return expected_.failed
        for index in expected_.checked_indices:
            self.entries[index].checked = True
        return None

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.uninstall()
        self.ensure_checked()


class LogCaptureForDecorator(LogCapture):
    def install(self) -> Self:
        LogCapture.install(self)
        self.clear()
        return self


def log_capture(*names: str, **kw: Any) -> Callable:
    """
    A decorator for making a :class:`LogCapture` installed and
    available for the duration of a test function.

    :param names: An optional sequence of names specifying the loggers
                  to be captured. If not specified, the root logger
                  will be captured.

    Keyword parameters other than ``install`` may also be supplied and will be
    passed on to the :class:`LogCapture` constructor.
    """
    l = LogCaptureForDecorator(names or None, install=False, **kw)
    return wrap(l.install, l.uninstall)


class LoggingHandler(logging.Handler):
    """Minimal handler that forwards records to a :class:`LoggingSource`."""

    def __init__(self, source: 'LoggingSource') -> None:
        super().__init__()
        self._source = source

    def __bool__(self) -> bool:
        # Some logging internals check boolean rather than identity for handlers :-(
        return True

    def emit(self, record: LogRecord) -> None:
        self._source._handle_record(record)

    def close(self) -> None:
        super().close()
        self._source._handler_closed()


class LoggingSource:
    """
    A :class:`~testfixtures.logcapture.CaptureSource` for the standard-library
    :mod:`logging` framework, for use with :class:`~testfixtures.LogCapture`.

    :param attributes:
        The sequence of attributes to return for each record or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, an attribute of that name will
        be obtained from the :class:`logging.LogRecord`. If the attribute is callable, the result of
        calling it will be used as the attribute value. If an attribute is missing, ``None`` will be
        used in its place.

        If a callable is passed, it will be called with the :class:`~logging.LogRecord` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.
    :param level:
        The minimum log level to capture.  Defaults to ``1``, capturing everything.
    :param names:
        A string (or tuple of strings) containing the dotted name(s) of loggers to capture.
        By default, the root logger is captured.
    :param propagate:
        If specified, any captured loggers will have their propagate attribute set to the supplied
        value. This can be used to prevent propagation from a child logger to a parent logger that
        has configured handlers.
    """

    def __init__(
        self,
        attributes: LogRecordAttributes = ('levelname', 'getMessage'),
        level: int = 1,
        *,
        names: Tuple[str | None, ...] = (None,),
        propagate: bool | None = None,
    ) -> None:
        self.attributes = attributes
        self.level = level
        self.names = names
        self.propagate = propagate
        self.old: dict[str, dict[str | None, Any]] = defaultdict(dict)
        self._collector: Callable[[Entry], None] | None = None
        self._handler: LoggingHandler | None = None
        self._compute_actual = build_actual_extractor(attributes, self.extract_field)

    def extract_field(self, raw: LogRecord, attribute: str) -> Any:
        value = getattr(raw, attribute, None)
        if callable(value):
            value = value()
        return value

    def __repr__(self) -> str:
        return f'LoggingSource({self.names!r})'

    def _handler_closed(self) -> None:
        if self._collector is not None:
            warn(
                f'LoggingSource closed while still capturing loggers: {self.names!r}\n'
                'Call uninstall() or use LogCapture as a context manager.'
            )

    def _handle_record(self, record: LogRecord) -> None:
        if self._collector is None:
            return
        exception = record.exc_info[1] if record.exc_info else None
        self._collector(Entry(
            raw=record,
            actual=self._compute_actual(record),
            level=record.levelno,
            exception=exception,
        ))

    def install(self, collector: Callable[[Entry], None]) -> None:
        self._collector = collector
        self._handler = LoggingHandler(self)
        for name in self.names:
            logger = logging.getLogger(name)
            self.old['levels'][name] = logger.level
            self.old['filters'][name] = logger.filters
            self.old['handlers'][name] = logger.handlers
            self.old['disabled'][name] = logger.disabled
            self.old['propagate'][name] = logger.propagate
            logger.setLevel(self.level)
            logger.filters = []
            logger.handlers = [self._handler]
            logger.disabled = False
            if self.propagate is not None:
                logger.propagate = self.propagate

    def uninstall(self) -> None:
        for name in self.names:
            logger = logging.getLogger(name)
            logger.setLevel(self.old['levels'][name])
            logger.filters = self.old['filters'][name]
            logger.handlers = self.old['handlers'][name]
            logger.disabled = self.old['disabled'][name]
            logger.propagate = self.old['propagate'][name]
        self._handler = None
        self._collector = None



from collections.abc import Iterable
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from functools import partial as partial_type
from inspect import signature
from pathlib import Path
from types import GeneratorType
from typing import (
    Any,
    List,
    Callable,
    overload,
    Self,
    TypeAlias,
    Literal,
    Iterator,
)
from typing import _GenericAlias as GenericAlias  # type: ignore[attr-defined]
from unittest.mock import call as unittest_mock_call

from testfixtures import not_there, singleton
from testfixtures.mock import mock_call
from .comparers import *

# Some common types that are immutable, for optimisation purposes within CompareContext
IMMUTABLE_TYPEs = str, bytes, int, float, tuple, type(None)

# Container types whose `__eq__` delegates to their elements. When per-type
# ignore_eq is in play, we can't trust `x == y` on these because nested
# instances of an ignored type would silently leak through.
CONTAINER_TYPES = list, tuple, dict, set, frozenset

# Things that are iterable, but should not be treated as such.
UNSAFE_ITERABLES = str, bytes, dict, GenericAlias


Comparer = Callable[[Any, Any, 'CompareContext'], str | None]
Comparers: TypeAlias = dict[type, Comparer]


DEFAULT_COMPARERS: Comparers = {
    dict: compare_dict,
    set: compare_set,
    frozenset: compare_set,
    list: compare_sequence,
    tuple: compare_tuple,
    str: compare_text,
    bytes: compare_bytes,
    int: compare_simple,
    float: compare_simple,
    Decimal: compare_simple,
    GeneratorType: compare_generator,
    mock_call.__class__: compare_call,
    unittest_mock_call.__class__: compare_call,
    BaseException: compare_exception,
    BaseExceptionGroup: compare_exception_group,
    partial_type: compare_partial,
    Path: compare_path,
    datetime: compare_with_fold,
    time: compare_with_fold,
}


@dataclass
class Registry:
    comparers: dict[type, Comparer]
    all_option_names: set[str]
    options_for: dict[Comparer, set[str]]
    ignore_eq_types: set[type]
    original: "Registry | None" = None

    @staticmethod
    def _shared_mro(x: Any, y: Any) -> Iterable[type]:
        y_mro = set(type(y).__mro__)
        for class_ in type(x).__mro__:
            if class_ in y_mro:
                yield class_

    def lookup(self, x: Any, y: Any, strict: bool) -> Comparer:
        if strict and type(x) is not type(y):
            return compare_with_type

        for class_ in self._shared_mro(x, y):
            comparer = self.comparers.get(class_)
            if comparer:
                return comparer

        # fallback for iterables
        if ((isinstance(x, Iterable) and isinstance(y, Iterable)) and not
            (isinstance(x, UNSAFE_ITERABLES) or
             isinstance(y, UNSAFE_ITERABLES))):
            return compare_generator

        return compare_object

    def __setitem__(self, key: type, value: Comparer) -> None:
        options = set(tuple(signature(value).parameters)[3:])
        self.options_for[value] = options
        self.all_option_names |= options
        self.comparers[key] = value

    @classmethod
    def initial(cls, comparers: Comparers | None = None) -> Self:
        registry = cls(
            comparers={},
            all_option_names = {'ignore_attributes'},
            options_for = {compare_object: {'ignore_attributes'}},
            ignore_eq_types = set(),
        )
        for name, value in (DEFAULT_COMPARERS if comparers is None else comparers).items():
            registry[name] = value
        return registry

    def copy(self) -> Self:
        return type(self)(
            comparers=self.comparers.copy(),
            all_option_names = self.all_option_names.copy(),
            options_for = self.options_for.copy(),
            ignore_eq_types = self.ignore_eq_types.copy(),
        )

    def overlay_with(self, comparers: Comparers) -> Self:
        registry = self.copy()
        for name, value in comparers.items():
            registry[name] = value
        return registry

    def install(self) -> Self:
        global _registry
        self.original = _registry
        _registry = self
        return self

    def uninstall(self) -> None:
        global _registry
        assert self.original is not None, "uninstall() called without install()"
        _registry = self.original
        del self.original

_registry = Registry.initial()


@contextmanager
def registry(comparers: Comparers | None = None) -> Iterator[Registry]:
    _registry = Registry.initial(comparers)
    _registry.install()
    try:
        yield _registry
    finally:
        _registry.uninstall()


@overload
def register(type_: type, comparer: Comparer) -> None: ...
@overload
def register(type_: type, comparer: Comparer, *, ignore_eq: bool) -> None: ...
@overload
def register(type_: type, *, ignore_eq: Literal[True]) -> None: ...


def register(
    type_: type,
    comparer: Comparer | None = None,
    *,
    ignore_eq: bool = False,
) -> None:
    """
    Register the supplied comparer for the specified type, and/or mark the
    type as one whose ``__eq__`` should be ignored during comparison.

    At least one of ``comparer`` or ``ignore_eq=True`` must be supplied.

    This registration is global and will be in effect from the point
    this function is called until the end of the current process.
    """
    if comparer is None and not ignore_eq:
        raise TypeError(
            "register() requires either a comparer or ignore_eq=True"
        )
    if comparer is not None:
        _registry[type_] = comparer
    if ignore_eq:
        _registry.ignore_eq_types.add(type_)


class CompareContext:
    """
    Stores the context of the current comparison in process during a call to
    :func:`testfixtures.compare`.
    """

    def __init__(
            self,
            x_label: str | None,
            y_label: str | None,
            recursive: bool = True,
            strict: bool = False,
            ignore_eq: bool | type | Iterable[type] = False,
            comparers: Comparers | None = None,
            options: dict[str, Any] | None = None,
    ):
        self._registry = _registry.overlay_with(comparers) if comparers else _registry
        if options:
            invalid = set(options) - self._registry.all_option_names
            if invalid:
                raise TypeError(
                    'The following options are not valid: ' + ', '.join(invalid)
                )
        self.x_label = x_label
        self.y_label = y_label
        self.recursive: bool = recursive
        self.strict: bool = strict
        self.ignore_eq_all: bool = False
        self.ignore_eq_types: set[type] = set(self._registry.ignore_eq_types)
        if ignore_eq is True:
            self.ignore_eq_all = True
        elif ignore_eq is False:
            pass
        elif isinstance(ignore_eq, type):
            self.ignore_eq_types.add(ignore_eq)
        else:
            self.ignore_eq_types.update(ignore_eq)
        self.options: dict[str, Any] = options or {}
        self.message: str = ''
        self.breadcrumbs: List[str] = []
        self._seen: dict[int, str] = {}

    def extract_args(self, args: tuple, x: Any, y: Any, expected: Any, actual: Any) -> List:

        possible = list[Any]()

        def append_if_specified(source: Any) -> None:
            if source is not unspecified:
                possible.append(source)

        append_if_specified(expected)
        possible.extend(args)
        append_if_specified(actual)
        append_if_specified(x)
        append_if_specified(y)

        if len(possible) != 2:
            message = 'Exactly two objects needed, you supplied:'
            if possible:
                message += ' {}'.format(possible)
            raise TypeError(message)

        return possible

    def label(self, side: Literal["x", "y"], value: Any) -> str:
        """
        Generate a labelled representation of the value for one side of a comparison.
        """
        r = str(value)
        label = getattr(self, side+'_label')
        if label:
            r += ' ('+label+')'
        return r

    def _separator(self) -> str:
        return '\n\nWhile comparing %s: ' % ''.join(self.breadcrumbs[1:])

    def _break_loops(self, obj: Any, breadcrumb: str) -> Any:
        # Don't bother with this process for simple, immutable types:
        if isinstance(obj, IMMUTABLE_TYPEs):
            return obj

        id_ = id(obj)
        breadcrumb_ = self._seen.get(id_)
        if breadcrumb_ is not None:
            return AlreadySeen(id_, obj, breadcrumb_)
        else:
            self._seen[id_] = breadcrumb
            return obj

    def qualified_equals(self, x: Any, y: Any) -> bool:
        """
        Determines if two objects are equal, taking ``strict``, ``ignore_eq`` and instances
        of :class:`~testfixtures.comparers.AlreadySeen` into account.
        """
        pair = x, y
        # When either side is an AlreadySeen wrapper for the other (same
        # underlying id), it's the same object we already handled —
        # equal by identity. Skipping __eq__ here keeps that guarantee
        # when __eq__ is broken, strict about unknown operands, or being
        # bypassed via ignore_eq.
        if any(type(o) is AlreadySeen for o in pair):
            a, b = (o.id if type(o) is AlreadySeen else id(o) for o in pair)
            if a == b:
                return True
        # AlreadySeen.__eq__ delegates to the wrapped object, so let
        # normal equality run when the wrapper is on the right.
        if type(y) is AlreadySeen:
            return x == y
        if self.strict or self.ignore_eq_all:
            return False
        # Containers delegate __eq__ to their elements, so when any
        # ignored type is in play we must block container == as well.
        types = self.ignore_eq_types
        if types and any(
            isinstance(o, CONTAINER_TYPES) or not types.isdisjoint(type(o).__mro__)
            for o in pair
        ):
            return False
        return x == y

    def call(self, comparer: Comparer, x: Any, y: Any) -> str | None:
        kw = {}
        option_names = self._registry.options_for.get(comparer)
        if option_names:
            for name in option_names:
                value = self.options.get(name, not_there)
                if value is not not_there:
                    kw[name] = value
        return comparer(x, y, self, **kw)

    def different(self, x: Any, y: Any, breadcrumb: str) -> bool | str | None:
        """
        Comparers can call this method to :ref:`hand off <custom-comparer-different>`
        comparison of elements within the objects they are currently comparing.
        """

        x = self._break_loops(x, breadcrumb)
        y = self._break_loops(y, breadcrumb)

        recursed = bool(self.breadcrumbs)
        self.breadcrumbs.append(breadcrumb)
        existing_message = self.message
        self.message = ''
        current_message = ''
        try:

            try:
                if self.qualified_equals(x, y):
                    return False
            except RecursionError:
                pass

            comparer: Comparer = self._registry.lookup(x, y, self.strict)

            result = self.call(comparer, x, y)
            specific_comparer = comparer is not compare_simple

            if result:

                if specific_comparer and recursed:
                    current_message = self._separator()

                if specific_comparer or not recursed:
                    current_message += result

                    if self.recursive:
                        current_message += self.message

            return result

        finally:
            self.message = existing_message + current_message
            self.breadcrumbs.pop()


def _resolve_lazy(source: Any) -> str:
    value = source() if callable(source) else source
    try:
        return str(value)
    except Exception:
        return safe_repr(value)


unspecified = singleton('unspecified')


def compare(
        *args: Any,
        x: Any = unspecified,
        y: Any = unspecified,
        expected: Any = unspecified,
        actual: Any = unspecified,
        prefix: str | Callable[[], str] | None = None,
        suffix: str | Callable[[], str] | None = None,
        x_label: str | None = None,
        y_label: str | None = None,
        raises: bool = True,
        recursive: bool = True,
        strict: bool = False,
        ignore_eq: bool | type | Iterable[type] = False,
        comparers: Comparers | None = None,
        **options: Any
) -> str | None:
    """
    Compare two objects, raising an :class:`AssertionError` if they are not
    the same. The :class:`AssertionError` raised will attempt to provide
    descriptions of the differences found.

    The two objects to compare can be passed either positionally or using
    explicit keyword arguments named ``x`` and ``y``, or ``expected`` and
    ``actual``, or a mixture of these.

    :param prefix: If provided, in the event of an :class:`AssertionError`
                   being raised, the prefix supplied will be prepended to the
                   message in the :class:`AssertionError`. This may be a
                   callable, in which case it will only be resolved if needed.

    :param suffix: If provided, in the event of an :class:`AssertionError`
                   being raised, the suffix supplied will be appended to the
                   message in the :class:`AssertionError`. This may be a
                   callable, in which case it will only be resolved if needed.

    :param x_label: If provided, in the event of an :class:`AssertionError`
                    being raised, the object passed as the first positional
                    argument, or ``x`` keyword argument, will be labelled
                    with this string in the message in the
                    :class:`AssertionError`.

    :param y_label: If provided, in the event of an :class:`AssertionError`
                    being raised, the object passed as the second positional
                    argument, or ``y`` keyword argument, will be labelled
                    with this string in the message in the
                    :class:`AssertionError`.

    :param raises: If ``False``, the message that would be raised in the
                   :class:`AssertionError` will be returned instead of the
                   exception being raised.

    :param recursive: If ``True``, when a difference is found in a
                      nested data structure, attempt to highlight the location
                      of the difference.

    :param strict: If ``True``, objects will only compare equal if they are
                   of the same type as well as being equal.

    :param ignore_eq: Controls when ``__eq__`` is skipped in favour of a
                      registered comparer (or the hash-equality fallback):

                      * ``True`` — skip ``__eq__`` for *every* object.
                      * ``False`` (default) — honour ``__eq__`` except for
                        types registered with ``ignore_eq=True``.
                      * a type — also skip ``__eq__`` whenever an instance of
                        that type is on either side of a comparison.
                      * an iterable of types — as above, for each type.

    :param comparers: If supplied, should be a dictionary mapping
                      types to comparer functions for those types. These will
                      be added to the comparer registry for the duration
                      of this call.

    Any other keyword parameters supplied will be passed to the functions
    that end up doing the comparison. See the
    :mod:`API documentation below <testfixtures.comparison>`
    for details of these.
    """

    __tracebackhide__ = True

    if not (expected is unspecified and actual is unspecified):
        x_label = x_label or 'expected'
        y_label = y_label or 'actual'

    context = CompareContext(x_label, y_label, recursive, strict, ignore_eq, comparers, options)
    x, y = context.extract_args(args, x, y, expected, actual)
    if not context.different(x, y, ''):
        return None

    message = context.message
    if prefix:
        message = _resolve_lazy(prefix) + ': ' + message
    if suffix:
        message += '\n' + _resolve_lazy(suffix)

    if raises:
        raise AssertionError(message)
    return message


try:
    from django.db.models import Model
    from .django import compare_model
except ImportError:
    pass
else:
    register(Model, compare_model, ignore_eq=True)


try:
    from pandas import DataFrame as PandasDataFrame
    from .pandas import compare_dataframe as compare_pandas_dataframe
except ImportError:
    pass
else:
    register(PandasDataFrame, compare_pandas_dataframe, ignore_eq=True)


try:
    from polars import DataFrame as PolarsDataFrame
    from .polars import compare_dataframe as compare_polars_dataframe
except ImportError:
    pass
else:
    register(PolarsDataFrame, compare_polars_dataframe, ignore_eq=True)


try:
    from numpy import ndarray
    from numpy.ma import MaskedArray
    from .numpy import compare_masked_array, compare_ndarray
except ImportError:
    pass
else:
    register(ndarray, compare_ndarray, ignore_eq=True)
    register(MaskedArray, compare_masked_array, ignore_eq=True)


try:
    from pydantic import BaseModel as PydanticBaseModel
    from .pydantic import compare_basemodel
except ImportError:
    pass
else:
    register(PydanticBaseModel, compare_basemodel, ignore_eq=True)



"""
Tools for testing entry-point scripts.
"""
import sys
from dataclasses import dataclass
from typing import Callable, Generic, Sequence, Self

from testfixtures.compat import TypeVar
from testfixtures.comparing import compare
from testfixtures.logcapture import LogCapture, LoggingSource
from testfixtures.mock import Mock, _Call as Call
from testfixtures.outputcapture import OutputCapture
from testfixtures.replace import Replacer


@dataclass
class CheckResult:
    """
    The result of a single check for one of the :class:`AbstractRun` attributes.
    Failure is indicated by providing a :attr:`message`.
    """

    #: The name of this check
    name: str
    #: The message describing why this check failed, or ``None`` if it passed.
    message: str | None


def check_output(expected: str, output: OutputCapture) -> CheckResult:
    """
    Check if the expected output matches the combined output written to :data:`~sys.stdout` and
    :data:`~sys.stderr` during a :meth:`~Command.run`.
    Used in the default :meth:`Run.check`.
    """
    return CheckResult('output', output.compare(expected, raises=False))


def check_logging(expected: Sequence[tuple[str, ...] | str], logging: LogCapture) -> CheckResult:
    """
    Check if the expected logging matches that :attr:`captured <AbstractRun.logging>` during a
    :meth:`~Command.run`.
    Used in the default :meth:`Run.check`.
    """
    return CheckResult('logging', logging.check(*expected, raises=False))


def check_mock_calls(expected: Sequence[Call], mocks: Mock) -> CheckResult:
    """
    Check if the expected mock calls match those :attr:`captured <AbstractRun.mocks>` during a
    :meth:`~Command.run`.
    Used in the default :meth:`Run.check`.
    """
    return CheckResult('mock calls', compare(expected, actual=mocks.mock_calls, raises=False))


def check_return_code(expected: int, actual: int) -> CheckResult:
    """
    Check if the expected return code matches the actual one from a :meth:`~Command.run`.
    Used in the default :meth:`Run.check`.
    """
    return CheckResult('return_code', compare(expected=expected, actual=actual, raises=False))


@dataclass
class AbstractRun:
    """
    A base class used for implementing a :meth:`Command.run` and its result.
    """

    #: The :class:`~testfixtures.OutputCapture` that was active during the run.
    output: OutputCapture
    #: The return code derived from any :class:`SystemExit` raised during the run,
    #: or ``0`` if none was raised.
    return_code: int
    #: The :class:`~testfixtures.LogCapture` that was active during the run.
    logging: LogCapture
    #: The mocks used during the run.
    mocks: Mock

    @classmethod
    def setup_output(cls) -> OutputCapture:
        """
        Instantiate the :class:`~testfixtures.OutputCapture` used for each run.
        Can be overriden to make different capturing choices, such as not capturing at all
        or capturing :data:`~sys.stdout` or :data:`~sys.stderr` separately.
        """
        return OutputCapture()

    @classmethod
    def setup_logging(cls) -> LogCapture:
        """
        Instantiate the :class:`~testfixtures.LogCapture` used for each run.
        Can be overriden to only capture logging at or above a certain level or if a different
        logging framework is in use, such as :mod:`loguru` or :mod:`structlog`.
        """
        return LogCapture(LoggingSource())

    @classmethod
    def setup_mocks(cls, replace: Replacer) -> Mock:
        """
        Use the provided :class:`~testfixtures.Replacer` to mock any required callables or objects,
        either where they would interfere with testing of the callable, or where you want to check
        that things are called with appropriate options given the command line specified in
        :data:`sys.argv`.

        Any mocks installed should be attributes of a root :class:`unittest.mock.Mock` instance,
        with that root mock being returned for later :meth:`checking <Run.check>`.
        """
        return Mock()

    @staticmethod
    def check_output(expected: str, output: OutputCapture) -> CheckResult:
        """
        Override to customise the checking of :attr:`output`.
        """
        return check_output(expected, output)

    @staticmethod
    def check_return_code(expected: int, actual: int) -> CheckResult:
        """
        Override to customise the checking of :attr:`return_code`.
        """
        return check_return_code(expected, actual)

    @staticmethod
    def check_logging(
        expected: Sequence[tuple[str, ...] | str], logging: LogCapture
    ) -> CheckResult:
        """
        Override to customise the checking of :attr:`logging`.
        """
        return check_logging(expected, logging)

    @staticmethod
    def check_mock_calls(expected: Sequence[Call], mocks: Mock) -> CheckResult:
        """
        Override to customise the checking of calls on :attr:`mocks`.
        """
        return check_mock_calls(expected, mocks)

    @classmethod
    def produce(cls, main: Callable[[], None], argv: list[str]) -> Self:
        """
        Produce a :class:`run <AbstractRun>` by setting up all required mocking and capturing
        and then calling the supplied ``main`` with :data:`sys.argv` set to the supplied ``argv``.
        """
        with (
            Replacer() as replace,
            cls.setup_output() as output,
            cls.setup_logging() as logging,
        ):
            mocks = cls.setup_mocks(replace)
            replace(
                sys.argv,
                container=sys,
                name='argv',
                replacement=[getattr(main, '__name__', 'main')] + argv,
            )
            return_code = 0
            try:
                main()
            except SystemExit as e:
                match e.code:
                    case str():
                        print(e.code, file=sys.stderr)
                        return_code = 1
                    case int():
                        return_code = e.code
                    case None:
                        return_code = 1
        return cls(output, return_code, logging, mocks)

    def check_results(self, *checks: CheckResult) -> None:
        """
        Accumulate output from the supplied :class:`CheckResult` instances and raise an
        :class:`AssertionError` if any of them failed.
        """
        __tracebackhide__ = True
        problems = [f"{check.name}: {check.message}" for check in checks if check.message]
        if problems:
            raise AssertionError('\n\n'.join(problems))


class Run(AbstractRun):
    """
    The result of a single :meth:`Command.run`.
    """

    def check(
        self,
        output: str = '',
        return_code: int = 0,
        logging: Sequence[tuple[str, str]] = (),
        mock_calls: Sequence[Call] = (),
    ) -> None:
        """
        Check the result of a :meth:`Command.run`.

        :param output: Expected output written to either :data:`~sys.stdout` or :data:`~sys.stderr`.

        :param return_code: Expected return code.

        :param logging: Expected log entries to be :meth:`checked <testfixtures.LogCapture.check>`.

        :param mock_calls: Expected :attr:`~unittest.mock.Mock.mock_calls`.
        """
        __tracebackhide__ = True
        self.check_results(
            self.check_output(output, self.output),
            self.check_return_code(return_code, self.return_code),
            self.check_logging(logging, self.logging),
            self.check_mock_calls(mock_calls, self.mocks),
        )


RunT = TypeVar('RunT', bound=AbstractRun, default=Run)


@dataclass
class Command(Generic[RunT]):
    """
    A test harness for running an entry-point or script callable with various mocks in place.
    These mocks provide a synthetic :data:`sys.argv`, capture output written to
    :data:`~sys.stdout` or :data:`~sys.stderr`, capture logging, and capture
    any setup calls that have been mocked along with any :class:`SystemExit` raised.
    """

    #: The entry-point or script callable to run.
    main: Callable[[], None]
    #: The :class:`runner <AbstractRun>` to use.
    runner: type[RunT] = Run  # type: ignore[assignment]

    def run(self, *argv: str) -> RunT:
        """
        Run :attr:`main` with :data:`sys.argv` set to
        ``[main.__name__, *argv]`` with :meth:`~AbstractRun.setup_output` and
        :meth:`~AbstractRun.setup_logging` captured and
        :meth:`~AbstractRun.setup_mocks` in place.

        Any :class:`SystemExit` raised by ``main`` is translated to
        :attr:`AbstractRun.return_code` as follows:

        - :class:`int` codes are passed through.
        - :class:`str` codes are printed to stderr and reported as ``1``
        - A bare :class:`SystemExit` is reported as ``1``.

        """
        return self.runner.produce(self.main, list(argv))

    def __call__(self, *argv: str) -> RunT:
        """
        Alternative way of calling :meth:`run`.
        """
        return self.run(*argv)
